// knopf.c
// AVR-MT128 Olimex 
#include <avr/io.h>			// Muss immer sein
#include <avr/interrupt.h> 	// Fuer Interrupts

#include "main.h" 	// Allgemeine Programmablauf
#include "knopf.h" 	// Knopf

static u8 key_state; // debounced and inverted key state:
                  // bit = 1: key pressed
static u8 key_press; // key press detect
static u8 key_rpt;  // key long press and repeat

void knopf_abfrage(void){

	static u8 ct0 = 0xFF, ct1 = 0xFF, rpt;
	
	u8 i, key_pin;

	key_pin = knopf_input();	// Lesen von Knopf-Zustand
	
	i = key_state ^ ~key_pin;     // key changed ?
	ct0 = ~( ct0 & i );           // reset or count ct0
	ct1 = ct0 ^ (ct1 & i);        // reset or count ct1
	i &= ct0 & ct1;               // count until roll over ?
	key_state ^= i;               // then toggle debounced state
	key_press |= key_state & i;   // 0->1: key press detect

	if( (key_state & REPEAT_MASK) == 0 )  // check repeat function
     rpt = REPEAT_START;                  // start delay
  	if( --rpt == 0 ){
    rpt = REPEAT_NEXT;                    // repeat delay
    key_rpt |= key_state & REPEAT_MASK;
  }
}

/* Pruefen Sie, ob eine Taste gedrueckt wurde. Jede gedrueckte Taste wird nur einmal gemeldet */
//
u8 get_key_press( u8 key_mask )
{
  u8 tmp_sreg;	// temporaerer Speicher fuer das Statusregister
  tmp_sreg = SREG; // Statusregister (also auch das I-Flag darin) sichern
  cli();                                          // read and clear atomic !

  key_mask &= key_press;                          // read key(s)
  key_press ^= key_mask;                          // clear key(s)
  SREG = tmp_sreg; // Status-Register wieder herstellen
     // somit auch das I-Flag auf gesicherten Zustand setzen
  return key_mask;
}

/* Pruefen Sie, ob eine Taste so lange gedrueckt wurde, dass die 
Tastenwiederholungsfunktion aktiviert wird. Nach einer kurzen 
Einstellungsverzoegerung wird gemeldet, dass die Taste bei 
nachfolgenden Aufrufen dieser Funktion gedrueckt wird. 
Dies simuliert, dass der Benutzer die Taste wiederholt drueckt und freigibt.
*/
u8 get_key_rpt( u8 key_mask ){
  u8 tmp_sreg;	// temporaerer Speicher fuer das Statusregister
  tmp_sreg = SREG; // Statusregister (also auch das I-Flag darin) sichern
  cli();                    // read and clear atomic !

  key_mask &= key_rpt;      // read key(s)
  key_rpt ^= key_mask;      // clear key(s)
  SREG = tmp_sreg; // Status-Register wieder herstellen
     // somit auch das I-Flag auf gesicherten Zustand setzen
  return key_mask;
}

/* Pruefen Sie, ob jetzt eine Taste gedrueckt wird */
//
u8 get_key_state( u8 key_mask ){
  key_mask &= key_state;
  return key_mask;
}

//
u8 get_key_short( u8 key_mask ){
	u8 x;
  	u8 tmp_sreg;	// temporaerer Speicher fuer das Statusregister
  	tmp_sreg = SREG; // Statusregister (also auch das I-Flag darin) sichern
  	cli();  // read key state and key press atomic !

  	x = get_key_press( ~key_state & key_mask );
  	SREG = tmp_sreg; // Status-Register wieder herstellen
     // somit auch das I-Flag auf gesicherten Zustand setzen
  	return x;
}

//
u8 get_key_long( u8 key_mask ){
  return get_key_press( get_key_rpt( key_mask ));
}

//u8 get_key_long_r( u8 key_mask ){ // if repeat function needed
//  return get_key_press( get_key_rpt( key_press & key_mask ));
//}


//u8 get_key_rpt_l( u8 key_mask ){ // if long function needed
//  return get_key_rpt( ~key_press & key_mask );
//}

/* Funktionsweise */
/* Der Code basiert auf 8 parallelen vertikalen Zaehlern, 
die ueber die Variablen ct0 und ct1 aufgebaut werden */
/* 8 vertikale Zaehler in 2 8-Bit Variablen */
/* wobei jeweils ein Bit in ct0 mit dem gleichwertigen Bit 
in ct1 zusammengenommen einen 2-Bit-Zaehler bildet. Der Code 
der sich um die 8 Zaehler kuemmert, ist so geschrieben, dass er 
alle 8 Zaehler gemeinsam parallel behandelt. */
//  i = key_state ^ ~KEY_PIN;               // key changed ?
/* i enthaelt an dieser Stelle fuer jede Taste, die sich im Vergleich 
mit dem vorhergehenden entprellten Zustand (keystate) veraendert hat, 
ein 1 Bit. */
// ct0 = ~( ct0 & i );                     // reset or count ct0
// ct1 = ct0 ^ (ct1 & i);                  // reset or count ct1
/* Diese beiden Anweisungen erniedrigen den 2-Bit Zaehler ct0/ct1 
fuer jedes Bit um 1, welches in i gesetzt ist. Liegt an der 
entsprechenden Stelle in i ein 0 Bit vor (keine Aenderung des Zustands), 
so wird der Zaehler ct0/ct1 fuer dieses Bit auf 1 gesetzt. 
Der Grundzustand des Zaehlers ist als ct0 == 1 und ct1 == 1 (Wert 3). 
Der Zaehler zaehlt daher mit jedem ISR Aufruf, bei dem die Taste 
im Vergleich zu keystate als veraendert erkannt wurde */
//   ct1   ct0
//    1    1   // 3
//    1    0   // 2
//    0    1   // 1
//    0    0   // 0
//    1    1   // 3
// i &= ct0 & ct1;       // count until roll over ?
/* in i bleibt nur dort ein 1-Bit erhalten, wo sowohl in ct1 als auch 
in ct0 ein 1 Bit vorgefunden wird, der betreffende Zaehler also bis 3 
zaehlen konnte. Durch die zusaetzliche Verundung mit i wird der Fall 
abgefangen, dass ein konstanter Zaehlerwert von 3 in i ein 1 Bit 
hinterlaesst. Im Endergebnis bedeutet dass, dass nur ein Zaehlerwechsel 
von 0 auf 3 zu einem 1 Bit an der betreffenden Stelle in i fuehrt, 
aber auch nur dann, wenn in i an dieser Bitposition ebenfalls 
ein 1 Bit war (welches wiederrum deswegen auf 1 war, weil an diesem 
Eingabeport eine Veraenderung zum letzten bekannten entprellten Zustand 
festgestellt wurde). Alles zusammengenommen heisst das, dass ein 
Tastendruck dann erkannt wird, wenn die Taste 4 mal hintereinander 
in einem anderen Zustand vorgefunden wurde als dem zuletzt bekannten 
entprellten Tastenzustand. */
/* An dieser Stelle ist i daher ein Vektor von 8 Bits, von denen 
jedes einzelne der Bits darueber Auskunft gibt, ob die entsprechende 
Taste mehrmals hintereinander im selben Zustand angetroffen wurde, 
der nicht mit dem zuletzt bekannten Tastenzustand uebereinstimmt. 
Ist das der Fall, dann wird eine entsprechende Veraenderung des 
Tastenzustands in key_state registriert */
// key_state ^= i;  // then toggle debounced state
/* und wenn sich in key_state das entsprechende Bit von 0 auf 1 
veraendert hat, wird dieses Ereignis als 'Taste wurde niedergedrueckt' 
gewertet. */
// key_press |= key_state & i;  // 0->1: key press detect
/* Damit ist der Tasteneingang entprellt. Und zwar sowohl beim Druecken 
einer Taste als auch beim Loslassen (damit Tastenpreller beim Loslassen 
nicht mit dem Niederdruecken einer Taste verwechselt werden). 
Der weitere Code beschaeftigt sich dann nur noch damit, diesen entprellten 
Tastenzustand weiter zu verarbeiten. */
/* Der Codeteil sieht durch die Verwendung der vielen bitweisen 
Operationen relativ komplex aus. Behaelt man aber im Hinterkopf, dass 
einige der bitweisen wie ein 'paralles If' gleichzeitig auf allen 8 Bits 
eingesetzt werden, dann vereinfacht sich vieles. Ein */
// key_press |= key_state & i;
/* ist nichts anderes als ein */
	  // teste ob Bit 0 sowohl in key_state als auch in i gesetzt ist
      // und setze Bit 0 in key_press, wenn das der Fall ist
// if( ( key_state & ( 1 << 0 ) ) && ( i & ( 1 << 0 ) )
//       key_press |= ( 1 << 0 );

// Bit 1
//    if( ( key_state & ( 1 << 1 ) ) && ( i & ( 1 << 1 ) )
//       key_press |= ( 1 << 1 );

// Bit 2
//    if( ( key_state & ( 1 << 2 ) ) && ( i & ( 1 << 2 ) )
//       key_press |= ( 1 << 2 );

 //   ...

/* nur als wesentlich kompaktere Operation ausgefuehrt und fuer alle 
8 Bits gleichzeitig. Die Kuerze und Effizienz dieser paar Codezeilen 
ergibt sich aus dem Umstand, dass jedes Bit in den Variablen fuer 
eine Taste steht und alle 8 (maximal moeglichen) Tasten gleichzeitig 
die Operationen durchlaufen. */
/* Funktionsweisen der verschienden Modi anhand von Zeitstrahlen 
erklaert: https://www.mikrocontroller.net/topic/48465?goto=1753367#1844458 */
/* "Walkthrough" der verschiedenen Zustaender der Einzelnen Variablen 
anhand eines Tastendrucks (avrfreaks.net) 
http://www.avrfreaks.net/comment/726676#comment-726676 */

/* Anmerkung:

Erlaubte Funktionen einer Taste:

- get_key_press()

- get_key_rpt()

- get_key_press() mit get_key_rpt()

- get_key_short() mit get_key_long()

- get_key_short() mit get_key_long_r() und get_key_rpt_l()

Peter */
