// lcd.c
// AVR-MT128 Olimex
#include <avr/io.h>			// Muss immer sein
#include <avr/pgmspace.h> 	// Fuer Flash 
#include <util/delay.h>		// Fuer Verzoegerungen

#include "main.h" 	// Allgemeine Programmablauf 
#include "lcd.h"	// LCD
#include "rtos.h"

// BUSY wird nicht benutzt, R/W immer 0

/*************** Variablen ************************************/
static u8 lcdpuff[LCD_PUFFER] = {}; 
static u8 lcdpuff_index = 0;
u8 lcd_erneuern = PUF_ERNEUERN;
	// wenn PUF_ERNEUERN (0), Beginn
	// wenn LCD_ZY_MAX erricht, Stopp 
/***********   *******************************/

static inline void lcd_enable_puls(void){
	LCD_PORT_EN |= (1<<LCD_EN);     // Enable auf 1 setzen
//    _delay_us( LCD_ENABLE_US );  // kurze Pause
	asm volatile("nop");
	asm volatile("nop");	// 6+2=8c, 62,5 nc * 8 = 500 nc > 450 nc min
	asm volatile("nop");
	asm volatile("nop");
	asm volatile("nop");
	asm volatile("nop");
    LCD_PORT_EN &= ~(1<<LCD_EN);    // Enable auf 0 setzen	
}

static inline void lcd_dataline_null(void){
	LCD_PORT_DB4 &= ~(1<<LCD_DB4); // initial alle Ausgaenge auf Null
	LCD_PORT_DB5 &= ~(1<<LCD_DB5);
	LCD_PORT_DB6 &= ~(1<<LCD_DB6);
	LCD_PORT_DB7 &= ~(1<<LCD_DB7);
}


// Sendet eine 4-bit Ausgabeoperation an das LCD 
static void lcd_out( u8 data ) {
	lcd_dataline_null(); 
	if (data & 0x10){
		LCD_PORT_DB4 |= (1<<LCD_DB4);
	}
	if (data & 0x20){
		LCD_PORT_DB5 |= (1<<LCD_DB5);
	}
	if (data & 0x40){
		LCD_PORT_DB6 |= (1<<LCD_DB6);
	}
	if (data & 0x80){
		LCD_PORT_DB7 |= (1<<LCD_DB7); 
	}
	lcd_enable_puls();
}
/***************************************************************/

// Initialisierung: muss ganz am Anfang des Programms aufgerufen werden.
void lcd_init( void ) {
// verwendete Pins auf Ausgang schalten
	LCD_PORT_RS &= ~(1<<LCD_RS);
	LCD_PORT_EN &= ~(1<<LCD_EN);
	LCD_PORT_RW &= ~(1<<LCD_RW);

	lcd_dataline_null();

	LCD_DDR_EN |= (1<<LCD_EN); // Als Ausgaenge
	LCD_DDR_RS |= (1<<LCD_RS);
	LCD_DDR_RW |= (1<<LCD_RW);
	LCD_DDR_DB4 |= (1<<LCD_DB4);
	LCD_DDR_DB5 |= (1<<LCD_DB5);
	LCD_DDR_DB6 |= (1<<LCD_DB6);
	LCD_DDR_DB7 |= (1<<LCD_DB7);

// warten auf die Bereitschaft des LCD
    _delay_ms( LCD_BOOTUP_MS ); 

// Soft-Reset muss 3mal hintereinander gesendet werden zur Initialisierung
    lcd_out( LCD_SOFT_RESET );
    _delay_ms( LCD_SOFT_RESET_MS1 );
	lcd_enable_puls();
    _delay_ms( LCD_SOFT_RESET_MS2 );
	lcd_enable_puls();
    _delay_ms( LCD_SOFT_RESET_MS3 );

// 4-bit Modus aktivieren 
    lcd_out( LCD_SET_FUNCTION |
             LCD_FUNCTION_4BIT );
    _delay_ms( LCD_SET_4BITMODE_MS );

// 4-bit Modus / 2 Zeilen / 5x7
    lcd_command( LCD_SET_FUNCTION |
                 LCD_FUNCTION_4BIT |
                 LCD_FUNCTION_2LINE |
                 LCD_FUNCTION_5X7 );

// Display ein / Cursor aus / Blinken aus
    lcd_command(LCD_SET_DISPLAY |
                LCD_DISPLAY_ON |
                LCD_CURSOR_OFF |
                LCD_BLINKING_OFF);

// Cursor inkrement / kein Scrollen
    lcd_command( LCD_SET_ENTRY |
                 LCD_ENTRY_INCREASE |
                 LCD_ENTRY_NOSHIFT );

	lcd_clear();
}

/*******************************************************/
/* Funktionen fuer "gewoenliche" Arbeit mit LCD */

// Sendet ein Datenbyte an das LCD
void lcd_data( u8 data ) {

	LCD_PORT_RS |= (1<<LCD_RS);    // RS auf 1 setzen
	lcd_out( data );            // zuerst die oberen, 
    lcd_out( data<<4 );         // dann die unteren 4 Bit senden

    _delay_us( LCD_WRITEDATA_US );
}

// Sendet einen Befehl an das LCD
void lcd_command( u8 data ) {

	LCD_PORT_RS &= ~(1<<LCD_RS);    // RS auf 0 setzen
	lcd_out( data );             // zuerst die oberen, 
    lcd_out( data<<4);           // dann die unteren 4 Bit senden

    _delay_us(LCD_COMMAND_US );
}

// Sendet den Befehl zur Loeschung des Displays
void lcd_clear( void ) {

    lcd_command( LCD_CLEAR_DISPLAY );
    _delay_ms( LCD_CLEAR_DISPLAY_MS );
}

// Sendet den Befehl: Cursor Home
void lcd_home( void ) {
	
    lcd_command( LCD_CURSOR_HOME );
    _delay_ms( LCD_CURSOR_HOME_MS );
}

// Setzt den Cursor in Zeile y (0..3) Spalte x (0..19)
void lcd_setcursor( u8 x, u8 y ) {
    u8 data;
 
    switch (y) {
        case 0:    // 1. Zeile
            data = LCD_SET_DDADR + LCD_DDADR_LINE1 + x;
            break;
 
        case 1:    // 2. Zeile
            data = LCD_SET_DDADR + LCD_DDADR_LINE2 + x;
            break;
 
        case 2:    // 3. Zeile
            data = LCD_SET_DDADR + LCD_DDADR_LINE3 + x;
            break;
 
        case 3:    // 4. Zeile
            data = LCD_SET_DDADR + LCD_DDADR_LINE4 + x;
            break;
 
        default:    
            return; // fuer den Fall einer falschen Zeile
    }
 
    lcd_command( data );
}

// Schreibt einen String auf das LCD
// Ausgabe eines Strings an der aktuellen Cursorposition 
// String liegt im RAM
void lcd_string( u8 *data ) {
    while( *data != '\0' )
        lcd_data( *data++ );
}

// Schreibt einen String auf das LCD
// String liegt direkt im Flash Speicher
//	lcd_string_P(PSTR("Ich bin ich!")); 
void lcd_string_P( PGM_P data ) {
    u8 tmp;

    tmp=pgm_read_byte(data);
    while( tmp != '\0' ) {        
        lcd_data( tmp );
        data++;
        tmp = pgm_read_byte(data);
    }
}

// Schreibt ein Zeichen in den Character Generator RAM
// Daten liegen direkt im RAM
void lcd_generatechar( u8 code, const u8 *data ) {
    u8 i;

    // Startposition des Zeichens einstellen
    lcd_command( LCD_SET_CGADR | (code<<3) );
    // Bitmuster uebertragen
    for ( i=0; i<8; i++ ) {
        lcd_data( *data);
        data++;
    }
}

// Schreibt ein Zeichen in den Character Generator RAM
// Daten liegen direkt im Flash-Speicher
void lcd_generatechar_P( u8 code, PGM_P data ) {
    u8 i;

    // Startposition des Zeichens einstellen
    lcd_command( LCD_SET_CGADR | (code<<3) );
    // Bitmuster uebertragen
    for ( i=0; i<8; i++ ) {
        lcd_data( pgm_read_byte(data) );
        data++;
    }
}

/*******************************************************/
/* Funktionen fuer Arbeit mit LCD-puffer */ 

/* Mit lcd_erneuern = PUF_ERNEUERN wird Automatik gestartet, die LCD-Puffer 
nach LCD Zeichen um Zeichen jede 1 ms kopiert */
/* Wenn DISPATCHER_ENABLE definiert, wird auch die Aufgabe (lcd_automat) aus main.c
gestartet (sie sie wird selbst ausgeschaltet, wenn lcd_erneuern LCD_ZY_MAX erreicht */

static inline u8 lcd_puff_adresse(u8 base, u8 index){
/* aus base und index wird faktische Adresse gerechnet */
	u8 yy;
	yy = base + index;
	if(yy >= LCD_PUFFER) yy -= LCD_PUFFER;
	return yy;
}

/*void lcd_pdat(uint8_t data, uint8_t platz){ // Schreibt data in lcd-puffer ohne Automat zu schalten
// Reines Schreiben in Puffer
// ist notwendig, um bei Zeilen nur einmal lcd-puffer-Automatik startete
	if(platz < LCD_PUFFER){
		lcdpuff[lcd_puff_adresse(lcdpuff_index,platz)] = data;
	}
}*/

void puf_neu(void){ // Setzt Zeichen fuer Puffer-Kopieren
	lcd_erneuern = PUF_ERNEUERN;	// Notwendigkeit, LCD zu schreiben
	#ifdef DISPATCHER_ENABLE
		SetTask(lcd_automat,0,0,0);	// startet Kopieren aus Puffer in LCD
	#endif
}

void lcd_puff_data(u8 data, u8 platz){ // Schreibt data in lcd-puffer

//	lcd_pdat(data, platz);
//	puf_neu();
	if(platz < LCD_PUFFER){
		lcdpuff[lcd_puff_adresse(lcdpuff_index,platz)] = data;
	}
}

void lcd_puff_null(void){ // LCD-Puffer = Leerzeichen	

	for(int x=0;x<LCD_PUFFER;x++){
		lcdpuff[x] = ' '; // Leerzeichen
	}
	puf_neu();
} 

/* String ueber LCD-Puffer */

// Schreibt einen String auf das Puffer-LCD
void lcd_puff_string( u8 *data, u8 anf_platz ) {
    while( *data != '\0' ){
		lcd_puff_data( *data++, anf_platz++);
	}
	puf_neu();
}

// Schreibt einen String auf das Puffer-LCD
// String liegt direkt im Flash Speicher
// lcd_puff_string_P(PSTR("Zeit: St,Mn,Sk"),0);
void lcd_puff_string_P( PGM_P data, u8 anf_platz ) {
    u8 tmp;

    tmp=pgm_read_byte(data);
    while( tmp != '\0' ) {        
   		lcd_puff_data( tmp, anf_platz);
        data++;
		anf_platz++;
        tmp = pgm_read_byte(data);
    }
	puf_neu();
}
/****************************/

/* Funktion fuer automatische Kopieren aus puffer nach LCD, wird jede 1 ms gemacht */
void lcd_puff_lcd(void){ // LCD-Puffer >> LCD, jede 1 ms
	// lcd_erneuern = 0 startet Kopieren von Puffer nach LCD
	u8 data = 0;
	u8 data_index = 0;

	if(lcd_erneuern < LCD_ZY_MAX){
		#ifdef LCD_ZEILEN_1
			switch(lcd_erneuern){
				case LCD_ZY_LINE1 :
					data = (LCD_SET_DDADR + LCD_DDADR_LINE1);	// LCD auf Line 1
					LCD_PORT_RS &= ~(1<<LCD_RS);	// Befehl
					break;
				
				default :
					LCD_PORT_RS |= (1<<LCD_RS);	// Daten
					data_index = lcd_erneuern - 1; 
			
			data = lcdpuff[lcd_puff_adresse(lcdpuff_index,data_index)];
			}
		#endif

		#ifdef LCD_ZEILEN_2
			switch(lcd_erneuern){
				case LCD_ZY_LINE1 :
				data = (LCD_SET_DDADR + LCD_DDADR_LINE1);	// LCD auf Line 1
				LCD_PORT_RS &= ~(1<<LCD_RS);	// Befehl
				break;
			
			case LCD_ZY_LINE2 :
				data = (LCD_SET_DDADR + LCD_DDADR_LINE2);	// LCD auf Line 2
				LCD_PORT_RS &= ~(1<<LCD_RS);	// Befehl
				break;

			default :
				LCD_PORT_RS |= (1<<LCD_RS);	// Daten
				if(lcd_erneuern < LCD_ZY_LINE2){
					data_index = lcd_erneuern - 1;
				} else {
					data_index = lcd_erneuern - 2;
				}
			
			data = lcdpuff[lcd_puff_adresse(lcdpuff_index,data_index)];
			}
		#endif

		#ifdef LCD_ZEILEN_4
			switch(lcd_erneuern){

			case LCD_ZY_LINE1 :
				data = (LCD_SET_DDADR + LCD_DDADR_LINE1);	// LCD auf Line 1
				LCD_PORT_RS &= ~(1<<LCD_RS);	// Befehl
				break;
			
			case LCD_ZY_LINE2 :
				data = (LCD_SET_DDADR + LCD_DDADR_LINE2);	// LCD auf Line 2
				LCD_PORT_RS &= ~(1<<LCD_RS);	// Befehl
				break;

			case LCD_ZY_LINE3 :
				data = (LCD_SET_DDADR + LCD_DDADR_LINE3);	// LCD auf Line 3
				LCD_PORT_RS &= ~(1<<LCD_RS);	// Befehl
				break;

			case LCD_ZY_LINE4 :
				data = (LCD_SET_DDADR + LCD_DDADR_LINE4);	// LCD auf Line 4
				LCD_PORT_RS &= ~(1<<LCD_RS);	// Befehl
				break;

			default :

				LCD_PORT_RS |= (1<<LCD_RS);	// Daten
				if(lcd_erneuern < LCD_ZY_LINE2){
					data_index = lcd_erneuern - 1;
				} else if(lcd_erneuern < LCD_ZY_LINE3) {
					data_index = lcd_erneuern - 2;
				} else if(lcd_erneuern < LCD_ZY_LINE4) {
					data_index = lcd_erneuern - 3;
				} else {
					data_index = lcd_erneuern - 4;
				}

			data = lcdpuff[lcd_puff_adresse(lcdpuff_index,data_index)];
			}
		#endif

		lcd_erneuern++;
		
		lcd_out(data);	// Auf LCD schreiben
		lcd_out(data << 4);
	}
}

#ifdef DISPATCHER_ENABLE
	void lcd_automat(void){	/* LCD-Puffer nach LCD kopieren */
		lcd_puff_lcd();
		if(lcd_erneuern < LCD_ZY_MAX){	/* Automat ausschalten, wenn zu Ende */
			SetTimerTask(lcd_automat,DISPATCHER_TIME,0,0,0);
		}
}
#endif
