// lcd.h
// AVR-MT128 Olimex
#ifndef LCD_H
#define LCD_H 1

#include "mein_def.h"

// BUSY wird nicht benutzt, R/W immer 0

/*************** Variablen ************************************/
extern u8 lcd_erneuern;
	// wenn PUF_ERNEUERN (0), Beginn
	// wenn LCD_ZY_MAX erricht, Stopp 

/****** Arbeit mit Dispatcher? **************/
#define DISPATCHER_ENABLE	1
#define DISPATCHER_TIME 3 // ms, Zeit zwischen Puffer > LCD - Kopiervorgaengen
/********************************************/

/****** Auswahl Zeilen von LCD **************/
//#define LCD_ZEILEN_1
#define LCD_ZEILEN_2
//#define LCD_ZEILEN_4
/********************************************/

/****** Auswahl Zeichen pro Zeile ***********/
#define LCD_ZEICHEN_PRO_ZEILE	16
/********************************************/

/************ Pins waehler ******************/
/* Alle Pins sind voellig frei zu waehlen ***/

#define LCD_PORT_DB4	PORTC
#define LCD_DDR_DB4		DDRC
#define LCD_PIN_DB4		PINC
#define LCD_DB4 		PC4

#define LCD_PORT_DB5	PORTC
#define LCD_DDR_DB5		DDRC
#define LCD_PIN_DB5		PINC
#define LCD_DB5			PC5

#define LCD_PORT_DB6	PORTC
#define LCD_DDR_DB6		DDRC
#define LCD_PIN_DB6		PINC
#define LCD_DB6			PC6

#define LCD_PORT_DB7	PORTC
#define LCD_DDR_DB7		DDRC
#define LCD_PIN_DB7		PINC
#define LCD_DB7			PC7

#define LCD_PORT_RS		PORTC
#define LCD_DDR_RS		DDRC
#define LCD_RS			PC0

#define LCD_PORT_EN		PORTC
#define LCD_DDR_EN		DDRC
#define LCD_EN			PC2

#define LCD_PORT_RW		PORTC
#define LCD_DDR_RW		DDRC
#define LCD_RW			PC1

//#define LCD_PORT_BL		PORTC
//#define LCD_DDR_BL		DDRC
//#define LCD_BL			PC 

/********************************************/

// Prototype ============================================================

// Initialisierung: muss ganz am Anfang des Programms aufgerufen werden.
void lcd_init( void );

/*******************************************************/
/***** Funktionen fuer "gewoenliche" Arbeit mit LCD ****/

// Ausgabe eines einzelnen Zeichens an der aktuellen Cursorposition 
void lcd_data( u8 data );

// Ausgabe eines Kommandos an das LCD.
void lcd_command( u8 data );

// LCD loeschen
void lcd_clear( void ); 

// Cursor in die 1. Zeile, 0-te Spalte
void lcd_home( void );

// Cursor an eine beliebige Position 
// Setzt den Cursor in Zeile y (0..3) Spalte x (0..19)
void lcd_setcursor( u8 spalte, u8 zeile );

// Ausgabe eines Strings an der aktuellen Cursorposition 
// String liegt im RAM
void lcd_string( u8 *data ); 


// Ausgabe eines Strings an der aktuellen Cursorposition
// String liegt im Flash 
void lcd_string_P( PGM_P data ); 

// Definition eines benutzerdefinierten Sonderzeichens.
// data muss auf ein Array[8] mit den Zeilencodes des zu definierenden Zeichens
// zeigen, Daten liegen im RAM
void lcd_generatechar( u8 code, const u8 *data ); 

// Definition eines benutzerdefinierten Sonderzeichens.
// data muss auf ein Array[8] mit den Zeilencodes des zu definierenden Zeichens
// zeigen, Daten liegen im FLASH
void lcd_generatechar_P( u8 code, PGM_P data );

/*******************************************************/
/********* Funktionen fuer Arbeit mit LCD-puffer *******/

#ifdef DISPATCHER_ENABLE
	void lcd_automat(void);	/* LCD-Puffer nach LCD kopieren */
#endif

/* Mit lcd_erneuern > 0 wird Automatik gestartet, die LCD-Puffer nach LCD
Zeichen um Zeichen jede 1 ms kopiert */
void lcd_puff_lcd(void); // LCD-Puffer >> LCD, jede 1 ms

void puf_neu(void); // Setzt Zeichen fuer Puffer-Kopieren

//void lcd_pdat(uint8_t data, uint8_t platz); // Schreibt data in lcd-puffer ohne Automat zu schalten

void lcd_puff_data(u8 data, u8 platz); // Schreibt data in lcd-puffer

void lcd_puff_null(void); // LCD-Puffer = 0x20 

// Schreibt einen String auf das Puffer-LCD
void lcd_puff_string( u8 *data, u8 anf_platz );

// Schreibt einen String auf das Puffer-LCD
// String liegt direkt im Flash Speicher
void lcd_puff_string_P( PGM_P data, u8 anf_platz );

/*******************************************************/

/*********** Konstanten ***************************/
// LCD Ausfuehrungszeiten (MS=Millisekunden, US=Mikrosekunden)

#define LCD_BOOTUP_MS           15
#define LCD_ENABLE_US           1
#define LCD_WRITEDATA_US        46
#define LCD_COMMAND_US          42
 
#define LCD_SOFT_RESET_MS1      5
#define LCD_SOFT_RESET_MS2      1
#define LCD_SOFT_RESET_MS3      1
#define LCD_SET_4BITMODE_MS     5
 
#define LCD_CLEAR_DISPLAY_MS    2
#define LCD_CURSOR_HOME_MS      2


#define LCD_DDADR_LINE1         0x00
#define LCD_DDADR_LINE2         0x40
#define LCD_DDADR_LINE3         (LCD_DDADR_LINE1 + LCD_ZEICHEN_PRO_ZEILE)
#define LCD_DDADR_LINE4         (LCD_DDADR_LINE2 + LCD_ZEICHEN_PRO_ZEILE) 

#ifdef LCD_ZEILEN_1
	#define LCD_ZY_LINE1	0
	#define LCD_ZY_MAX		(LCD_ZEICHEN_PRO_ZEILE + 1)
	// fuer Operationen mit LCD-Puffer
	#define LCD_ZEILEN 1
#endif

#ifdef LCD_ZEILEN_2
	#define LCD_ZY_LINE1	0
	#define LCD_ZY_LINE2	(LCD_ZEICHEN_PRO_ZEILE + 1)
	#define LCD_ZY_MAX		(LCD_ZY_LINE2 * 2)
	// fuer Operationen mit LCD-Puffer
	#define LCD_ZEILEN 2
#endif

#ifdef LCD_ZEILEN_4
	#define LCD_ZY_LINE1	0
	#define LCD_ZY_LINE2	(LCD_ZEICHEN_PRO_ZEILE + 1)
	#define LCD_ZY_LINE3	(LCD_ZY_LINE2 * 2)
	#define LCD_ZY_LINE4	(LCD_ZY_LINE2 * 3)
	#define LCD_ZY_MAX	(LCD_ZY_LINE2 * 4)
	// fuer Operationen mit LCD-Puffer
	#define LCD_ZEILEN 4
#endif

// fuer Puffergroesse
#define LCD_PUFFER (LCD_ZEICHEN_PRO_ZEILE * LCD_ZEILEN)

#define PUF_ERNEUERN 	0 

////////////////////////////////////////////////////////////////////////////////
// LCD Befehle und Argumente.
// Zur Verwendung in lcd_command
 
// Clear Display -------------- 0b00000001
#define LCD_CLEAR_DISPLAY       0x01
 
// Cursor Home ---------------- 0b0000001x
#define LCD_CURSOR_HOME         0x02
 
// Set Entry Mode ------------- 0b000001xx
#define LCD_SET_ENTRY           0x04
 
#define LCD_ENTRY_DECREASE      0x00
#define LCD_ENTRY_INCREASE      0x02
#define LCD_ENTRY_NOSHIFT       0x00
#define LCD_ENTRY_SHIFT         0x01
 
// Set Display ---------------- 0b00001xxx
#define LCD_SET_DISPLAY         0x08
 
#define LCD_DISPLAY_OFF         0x00
#define LCD_DISPLAY_ON          0x04
#define LCD_CURSOR_OFF          0x00
#define LCD_CURSOR_ON           0x02
#define LCD_BLINKING_OFF        0x00
#define LCD_BLINKING_ON         0x01
 
// Set Shift ------------------ 0b0001xxxx
#define LCD_SET_SHIFT           0x10
 
#define LCD_CURSOR_MOVE         0x00
#define LCD_DISPLAY_SHIFT       0x08
#define LCD_SHIFT_LEFT          0x00
//#define LCD_SHIFT_RIGHT         0x01
#define LCD_SHIFT_RIGHT         0x04
 
// Set Function --------------- 0b001xxxxx
#define LCD_SET_FUNCTION        0x20
 
#define LCD_FUNCTION_4BIT       0x00
#define LCD_FUNCTION_8BIT       0x10
#define LCD_FUNCTION_1LINE      0x00
#define LCD_FUNCTION_2LINE      0x08
#define LCD_FUNCTION_5X7        0x00
#define LCD_FUNCTION_5X10       0x04
 
#define LCD_SOFT_RESET          0x30
 
// Set CG RAM Address --------- 0b01xxxxxx  (Character Generator RAM)
#define LCD_SET_CGADR           0x40
 
#define LCD_GC_CHAR0            0
#define LCD_GC_CHAR1            1
#define LCD_GC_CHAR2            2
#define LCD_GC_CHAR3            3
#define LCD_GC_CHAR4            4
#define LCD_GC_CHAR5            5
#define LCD_GC_CHAR6            6
#define LCD_GC_CHAR7            7
 
// Set DD RAM Address --------- 0b1xxxxxxx  (Display Data RAM)
#define LCD_SET_DDADR           0x80

#endif	/* LCD_H */
