// main.c
// AVR-MT128 Olimex
// RTOS nach DI_HALT http://easyelectronics.ru
// benutzt typedef aus mein_def.h, Ueberschaubarkeit optimisiert
// Ringpuffer fuer TaskQueue
// die Aufgaben bekommen je 3 Parameter: u16,s16,u8
// Timer 3 macht Klang mit OC3B und OC3C, FastPWM, OCR3A = Frequenz, OCR3B und OCR3C bestimmen Verhaeltnis
#include <avr/io.h>	// Muss immer sein
#include <avr/interrupt.h> 	// Fuer Interrupts
#include <avr/pgmspace.h> 	// Fuer Flash

#include "main.h" 	// Allgemeine Programmablauf
#include "rtos.h"
#include "knopf.h"
#include "lcd.h"
#include "logik.h"
#include "schall.h"

/*********** Aufgaben Variablen ***************************/

/*********************************************************/

static inline void init_all(void){
// init_port 
	RELAY_PORT &= ~(1<<RELAY_);
	RELAY_DDR |= (1<<RELAY_);
// init_indikator 
 	lcd_init();
 	lcd_puff_null();
// init_schall
 	schall_init();
	knopf_init();
}

ISR(TIMER2_COMP_vect){
	TimerService();
}

int main(void){
	init_all();	//  
	InitRTOS();			// Kern initialisieren
	RunRTOS();			// Kern starten. 

// Hintergrundaufgaben starten
	SetTask(knopf_abf,0,0,0);	// Tastenabfrage starten
	SetTimerTask(logik_k,(KNOPF_TIME/2),0,0,0);	// Logik starten

while(1){
	// wdt_reset();	// Watchdog auf Null
	TaskManager();	// Dispatcher ausfuehren
}
	return 0;
} 

//============================================================================
// Aufgaben 

void knopf_abf(void){	/* Tastenabfrage starten */
	knopf_abfrage();
	SetTimerTask(knopf_abf,KNOPF_TIME,0,0,0);
}

void logik_k(void){	/* Tastenanalyse, 10 ms (KNOPF_TIME) */
	logik_controlle();
	SetTimerTask(logik_k,KNOPF_TIME,0,0,0);
} 

void schall_ein_knopf(void){	// PA_puff.par0 = u16 freq, PA_puff.par1 = s16 time
	schall_freq_einstellung(PA_puff.par0);
	SetTimerTask(schall_aus,(u16)PA_puff.par1,0,0,0);
}

void schall_aus(void){
	SCHALL_AUS;
}

void logik_zahlen(void){	// PA_puff.par0 = Variable, PA_puff.par1 = Wert zu addieren
	switch(PA_puff.par0){
		case 0 : kn_00 += PA_puff.par1; ind_dec5(kn_00,0); break;	// 00
		case 1 : kn_01 += PA_puff.par1; ind_dec5(kn_01,6); break;	// 01
		case 2 : kn_02 += PA_puff.par1; ind_dec5(kn_02,12); break;	// 02
		case 3 : kn_03 += PA_puff.par1; ind_dec5(kn_03,18); break;	// 03
		case 4 : kn_04 += PA_puff.par1; ind_dec5(kn_04,24);break;	// 04
	}
}
